#include <pthread.h>        // thread stuff
#include <assert.h>         // assert()
#include <strings.h>        // bzero()
#include <stdlib.h>         // malloc(), realloc(), etc
#include <stdio.h>          // fprintf()

#include "buff.h"

buff_t *
buff_Create ()
{
  buff_t * buff;
  
  buff = (buff_t *)malloc(sizeof(buff_t));
  assert(buff != NULL);

  buff->init = 0;
  
  return buff;
}

void
buff_Init (buff, max)
  buff_t * buff;
  int max;
{
  int status = 0;
  
  assert(buff != NULL);

  buff->init = 1;
  buff->leng = 0;
  buff->size = 0;
  buff->maxm = max;
  buff->data = NULL;
  
  bzero(&buff->mutex, sizeof(pthread_mutex_t));
  
  status = pthread_mutex_init(&buff->mutex, NULL);
  assert(status == 0);
}

void
buff_ClearAll (buff)
  buff_t * buff;
{
  int status = 0;
  
  assert(buff != NULL);
  
  if (buff->init != 1)
    assert(buff->init == 1);
  
  if (0 != (status = pthread_mutex_lock(&buff->mutex))) {
    fprintf(stderr, "buff_Append.. failed to get mutex lock, %d\n", status);
    return;
  }

  buff->leng = 0;
  bzero(buff->data, buff->size);
  
  if (0 != (status = pthread_mutex_unlock(&buff->mutex))) {
    fprintf(stderr, "buff_Append.. failed to unlock mutex, %d\n", status);
    return;
  }
}

int
buff_Append (buff, data, leng)
  buff_t * buff;
  void * data;
  int leng;
{
  int status = 0;
  
  assert(buff != NULL);
  assert(data != NULL);

  if (buff->init != 1)
    assert(buff->init == 1);

  if (leng <= 0)
    return 0;
  
  if (0 != (status = pthread_mutex_lock(&buff->mutex))) {
    fprintf(stderr, "buff_Append.. failed to get mutex lock, %d\n", status);
    return -1;
  }
  
  if (buff->maxm != BUFF_UNLIMITED && buff->maxm - buff->leng < leng) {
    fprintf(stderr, "buff_Append.. insufficient space in buffer\n");
    return -1;
  }
  
  if (buff->data == NULL) {
    buff->size = 1.5 * leng;
    
    if (buff->maxm != BUFF_UNLIMITED)
      buff->size = buff->size > buff->maxm ? buff->maxm : buff->size;
    
    buff->data = (void *)malloc(buff->size);
  }
  else if (buff->size - buff->leng < leng) {
    buff->size = 1.5 * (buff->size + leng);
    
    if (buff->maxm != BUFF_UNLIMITED)
      buff->size = buff->size > buff->maxm ? buff->maxm : buff->size;
    
    buff->data = (void *)realloc(buff->data, buff->size);
  }
  
  assert(buff->data != NULL);
  
  memcpy(buff->data + buff->leng, data, leng);
  buff->leng += leng;
  
  if (0 != (status = pthread_mutex_unlock(&buff->mutex))) {
    fprintf(stderr, "buff_Append.. failed to unlock mutex, %d\n", status);
    return -1;
  }
  
  return leng;
}

// although it is not strictly necessary to synchronize for this fuction,
// as all we're doing is returning one int from the buffer, and not modifying
// a thing -- however, it is possible that another thread might be actively
// appending data, or removing data - and thus the value that we're interested
// in (buff->maxm - buff->leng) could change quickly. and as we're interested
// in accurate data, getting a lock will provide a greater assurance of that.
//
int
buff_Remaining (buff)
  buff_t * buff;
{
  int leng = 0;
  int status = 0;
  
  assert(buff != NULL);
  
  if (buff->init != 1)
    assert(buff->init == 1);
  
  if (0 != (status = pthread_mutex_lock(&buff->mutex))) {
    fprintf(stderr, "buff_Remaining.. failed to get mutex lock, %d\n", status);
    return -1;
  }

  leng = buff->maxm - buff->leng;
  
  if (0 != (status = pthread_mutex_unlock(&buff->mutex))) {
    fprintf(stderr, "buff_Remaining.. failed to unlock mutex, %d\n", status);
    return -1;
  }
  
  return leng;
}

int
buff_ReadLine (buff)
  buff_t * buff;
{
  int i = 0;
  int status = 0;
  int result = -1;
  
  assert(buff != NULL);

  if (buff->init != 1)
    assert(buff->init == 1);
  
  if (0 != (status = pthread_mutex_lock(&buff->mutex))) {
    fprintf(stderr, "buff_ReadLine.. failed to get mutex lock, %d\n", status);
    return -1;
  }
  
  for (i = 0; i < buff->leng - 1; i++)
    if (*((char *)buff->data + i) == '\r' && *((char *)buff->data + i + 1) == '\n') {
      result = i + 2;
      break;
    }

  if (0 != (status = pthread_mutex_unlock(&buff->mutex))) {
    fprintf(stderr, "buff_ReadLine.. failed to unlock mutex, %d\n", status);
    return -1;
  }
  
  return result;
}

int
buff_IndexOf (buff, str, strlen, offs)
  buff_t * buff;
  char * str;
  int strlen;
  int offs;
{
  int i = 0;
  int j = 0;
  int status = 0;
  int result = -1;
  
  assert(buff != NULL);
  assert(str != NULL);
  assert(strlen > 0);
  assert(offs >= 0);

  if (buff->init != 1)
    assert(buff->init == 1);
  
  if (0 != (status = pthread_mutex_lock(&buff->mutex))) {
    fprintf(stderr, "buff_IndexOf.. failed to get mutex lock, %d\n", status);
    return -1;
  }
  
  fprintf(stderr, "buff_IndexOf.. '");
  fwrite(buff->data, 1, 20, stdout);
  fflush(stdout);
  fprintf(stderr, "'\n");
         
  for (i = offs; i < buff->leng - strlen; i++) {
    for (j = 0, status = 0; j < strlen; j++, status++) {
      if (*((char *)buff->data + i) != *((char *)str + j))
        break;
    }

    if (status == strlen) {
      result = i;
      break;
    }
  }
      
  if (0 != (status = pthread_mutex_unlock(&buff->mutex))) {
    fprintf(stderr, "buff_ReadLine.. failed to unlock mutex, %d\n", status);
    return -1;
  }
  
  return result;
}

int
buff_Substring (buff, dest, offs, leng)
  buff_t * buff;
  char * dest;
  int offs;
  int leng;
{
  int status = 0;
  
  assert(buff != NULL);
//assert(dest != NULL);
  assert(offs >= 0);
  assert(leng >= 1);

  if (buff->init != 1)
    assert(buff->init == 1);
  
  if (0 != (status = pthread_mutex_lock(&buff->mutex))) {
    fprintf(stderr, "buff_Substring.. failed to get mutex lock, %d\n", status);
    return -1;
  }
  
//fprintf(stdout, "\n\nbuff_Substring.. BEFORE [buff->leng]: '");
//fwrite(buff->data, 1, buff->leng, stdout);
//fflush(stdout);
//fprintf(stdout, "'\n\n");
  
  if (offs > buff->leng)
    return 0;
  else if (offs + leng > buff->leng)
    leng = buff->leng - offs;
  
  // copy the desired data into dest, if dest is not null. there may
  // be an instance where one simply wishes to remove a set amount of
  // data from the buffer, and by passing in a NULL pointer for dest,
  // all this, and more can be yours! otherwise, just copy the data
  if (dest != NULL)
    memcpy(dest, buff->data + offs, leng);

  // shift all of the remaining data in the buffer, "down" to fill in the gap
  //
  memcpy(buff->data + offs, buff->data + offs + leng, buff->leng - (offs + leng));
  buff->leng -= leng;

//fprintf(stdout, "\n\nbuff_Substring.. AFTER [buff->leng]: '");
//fwrite(buff->data, 1, buff->leng, stdout);
//fflush(stdout);
//fprintf(stdout, "'\n\n");
  
  
  if (0 != (status = pthread_mutex_unlock(&buff->mutex))) {
    fprintf(stderr, "buff_Substring.. failed to unlock mutex, %d\n", status);
    return -1;
  }

  return dest == NULL ? 0 : leng;
}

void
buff_Destroy (buff)
  buff_t * buff;
{
  assert(buff != NULL);
  
  if (buff->data != NULL)
    free(buff->data);
  
  buff->leng = 0;
  buff->size = 0;
  buff->maxm = 0;
  buff->data = NULL;
  
  pthread_mutex_destroy(&buff->mutex);
  
  free(buff);
}
