#include <pthread.h>        // for threading stuff
#include <assert.h>         // assert()
#include <stdio.h>          // fprintf()
#include <stdlib.h>         // malloc()
#include <strings.h>        // bzero()

#include "hash.h"

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Public Hash Methods
#endif

hash_t *
hash_Create ()
{
  hash_t * hash;
  
  hash = (hash_t *)malloc(sizeof(hash_t));
  assert(hash != NULL);

  bzero(hash, sizeof(hash_t));
  
  return hash;
}

void
hash_Init (hash)
  hash_t * hash;
{
  int status;
  
  assert(hash != NULL);
  
  status = pthread_mutex_init(&hash->mutex, NULL);
  assert(status == 0);
  
  hash->leng = 0;
  hash->size = 5381;
  hash->list = NULL;
  
  hash->list = (hash_bucket_t **)malloc(sizeof(hash_bucket_t *) * hash->size);
  assert(hash->list != NULL);
}

inline unsigned long
hash_Hash (hash, name, leng)
  hash_t * hash;
  unsigned const char * name;
  unsigned long leng;
{
  unsigned long valu;
  int byte;
  int pos;
  
  assert(hash != NULL);
  assert(name != NULL);

  valu = hash->size;
  byte = 0;
  pos  = 0;
  
  //
  // shamelessly stolen from: http://www.cs.yorku.ca/~oz/hash.html
  //
  while (byte = *name++)
  {
    valu = ((valu << 5) + valu) + byte; /* hash * 33 + c */
    
    if (++pos >= leng)
      break;
  }

  return valu % hash->size;
}

hash_key_t *
hash_Put (hash, name, nleng, data, dleng)
  hash_t * hash;
  unsigned const char * name;
  unsigned long nleng;
  unsigned const char * data;
  unsigned long dleng;
{
  hash_key_t * key;
  hash_bucket_t * bucket;
  unsigned long valu;
  
  assert(hash != NULL);
  assert(name != NULL);
  assert(data != NULL);

  // get the hash value for the data, to be used as its index
  valu = hash_Hash(hash, name, nleng);

  // if a bucket does not already exist in this location, then
  // create a new bucket, and put it in that index
  if (NULL == (bucket = hash->list[valu])) {
    bucket = hash_bucket_Create();
    hash_bucket_Init(bucket);
    hash->list[valu] = bucket;
  }

  // append the data to the bucket, which returns the key that
  // was used to store the data. this key will either be a newly
  // created key, if no value previously existed with this value,
  // or an existing key that matches the data being added.
  key = hash_bucket_AppendValue(bucket, name, nleng, data, dleng);
  
  return key;
}

hash_key_t *
hash_Get (hash, name, nleng)
  hash_t * hash;
  unsigned const char * name;
  unsigned long nleng;
{
  unsigned long valu;
  hash_key_t * key;
  hash_bucket_t * bucket;
  
  assert(hash != NULL);
  assert(name != NULL);

  valu = hash_Hash(hash, name, nleng);

  if (NULL == (bucket = hash->list[valu]))
    return NULL;
  
  if (NULL == (key = hash_bucket_GetByValue(bucket, name, nleng)))
    return NULL;
  
  return key;
}

void
hash_Remove (hash, name, nleng)
  hash_t * hash;
  unsigned const char * name;
  unsigned long nleng;
{
  unsigned long valu;
  hash_bucket_t * bucket;
  
  assert(hash != NULL);
  assert(name != NULL);
  
  valu = hash_Hash(hash, name, nleng);
  
  // the bucket which would hold this value, if it did exist, does
  // not exist itself, so, we're done already.
  if (NULL == (bucket = hash->list[valu]))
    return;
  
  hash_bucket_RemoveByValue(bucket, name, nleng);
}

// list
// delete

void
hash_Destroy (hash)
  hash_t * hash;
{
  int i = 0;
  
  assert(hash != NULL);

  if (hash->list != NULL) {
    for (i = 0; i < hash->size; i++)
      if (hash->list[i] != NULL)
        hash_bucket_Destroy(hash->list[i]);
    
    free(hash->list);
  }
  
  pthread_mutex_destroy(&hash->mutex);

  hash->leng = 0;
  hash->size = 0;
  hash->list = NULL;
  
  free(hash);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Private Bucket Methods
#endif

hash_bucket_t *
  hash_bucket_Create ()
{
  hash_bucket_t * bucket;
  
  bucket = (hash_bucket_t *)malloc(sizeof(hash_bucket_t));
  assert(bucket != NULL);
  
  bzero(bucket, sizeof(hash_bucket_t));
  
  return bucket;
}

void
hash_bucket_Init (bucket)
  hash_bucket_t * bucket;
{
  assert(bucket != NULL);
  
  bucket->leng = 0;
  bucket->size = 0;
  bucket->list = NULL;
}

hash_key_t *
hash_bucket_AppendValue (bucket, name, nleng, data, dleng)
  hash_bucket_t * bucket;
  unsigned const char * name;
  unsigned long nleng;
  unsigned const char * data;
  unsigned long dleng;
{
  int i = 0;
  hash_key_t * key = NULL;
  
  assert(bucket != NULL);
  assert(name   != NULL);
  assert(data   != NULL);
  assert(nleng  >= 1);
  assert(dleng  >= 1);
  
  // look through the list keys in this bucket. if the index is null,
  // skip on to the next one. otherwise, compare it to the given data
  // and see if it is the same. if it is the same, we exit the loop,
  // and do not bother creating a new key. if we finish going through
  // the list and find that this value has not been used before, then
  // we have to create a new key to hold it ... and add it to then we
  // add it to the bucket.
  for (i = 0; i < bucket->size && key == NULL; i++) {
    if (bucket->list[i] == NULL)
      continue;
    
    if (0 == hash_key_CompareToValue(bucket->list[i], name, nleng))
      key = bucket->list[i];
  }
  
  // if key is not null at this point, that is because we found an
  // existing key in this bucket that matches the data given, thus
  // we needn't do anything more than return this key. presto.
  if (key != NULL) {
    hash_key_SetValue(key, data, dleng);
    return key;
  }
  
  // create a new key to hold this data, because no key presently
  // exists with this value.
  key = hash_key_Create();
  hash_key_Init(key, name, nleng, data, dleng);

  // if there are no extra "slots" available in the bucket to hold the
  // key, then expand the size of the bucket accordingly.
  if (bucket->leng == bucket->size) {
    if (bucket->size <= 0) {
      bucket->size = 5;
      bucket->list = (hash_key_t **)malloc(sizeof(hash_key_t *) * bucket->size);
      bzero(bucket->list, sizeof(hash_key_t *) * bucket->size);
    }
    else {
      int oldsize = bucket->size;
      bucket->size *= 1.5;
      bucket->list  = (hash_key_t **)realloc(bucket->list, sizeof(hash_key_t *) * bucket->size);
      bzero(bucket->list + oldsize, (sizeof(hash_key_t *) * bucket->size) - oldsize);
    }
    
    assert(bucket->list != NULL);
  }

  // find the next empty slot in the bucket, and place this key there
  for (i = 0; i < bucket->size; i++)
    if (bucket->list[i] == NULL) {
      bucket->list[i] = key;
      bucket->leng++;
      break;
    }

  return key;
}

hash_key_t *
hash_bucket_GetByValue (bucket, name, nleng)
  hash_bucket_t * bucket;
  unsigned const char * name;
  unsigned long nleng;
{
  hash_key_t * key = NULL;
  int i = 0;
  
  assert(bucket != NULL);
  assert(name   != NULL);
  assert(nleng  >= 1);
  
  // for each "slot" in the bucket... if the slot is null, then skip on to the
  // next one. if the slot is not empty, then compare it to the provided data
  // and see if we have a match. if we do, then assign it to the key pointer,
  // which will subsequently terminate the loop, and then we'll return that key.
  for (i = 0; i < bucket->size && key == NULL; i++) {
    if (bucket->list[i] == NULL)
      continue;
    
    if (0 == hash_key_CompareToValue(bucket->list[i], name, nleng))
      key = bucket->list[i];
  }
  
  return key;
}

void
hash_bucket_RemoveByValue (bucket, name, nleng)
  hash_bucket_t * bucket;
  unsigned const char * name;
  unsigned long nleng;
{
  hash_key_t * key = NULL;
  int i = 0;
  
  assert(bucket != NULL);
  assert(name   != NULL);
  
  for (i = 0; i < bucket->size && key == NULL; i++) {
    if (bucket->list[i] == NULL)
      continue;
    
    if (0 == hash_key_CompareToValue(bucket->list[i], name, nleng)) {
      key = bucket->list[i];
      bucket->list[i] = NULL;
    }
  }
  
  if (key != NULL)
    hash_key_Destroy(key);
}

void
hash_bucket_Destroy (bucket)
  hash_bucket_t * bucket;
{
  int i = 0;
  
  assert(bucket != NULL);

  if (bucket->list != NULL) {
    for (i = 0; i < bucket->size; i++)
      if (bucket->list[i] != NULL) {
        hash_key_Destroy(bucket->list[i]);
        bucket->list[i] = NULL;
      }
        
    free(bucket->list);
  }
  
  pthread_mutex_destroy(&bucket->mutex);

  bucket->leng = 0;
  bucket->size = 0;
  bucket->list = NULL;
  
  free(bucket);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Private Key Methods
#endif

hash_key_t *
hash_key_Create ()
{
  hash_key_t * key;
  
  key = (hash_key_t *)malloc(sizeof(hash_key_t));
  assert(key != NULL);

  bzero(key, sizeof(hash_key_t));
  
  return key;
}

void
hash_key_Init (key, name, nleng, data, dleng)
  hash_key_t * key;
  unsigned const char * name;
  unsigned long nleng;
  unsigned const char * data;
  unsigned long dleng;
{
  assert(key  != NULL);
  assert(name != NULL);
  assert(data != NULL);
  
  key->dleng = dleng;
  key->nleng = nleng;

  key->name = (unsigned char *)malloc(key->nleng + 1);
  key->data = (unsigned char *)malloc(key->dleng + 1);
  
  assert(key->data != NULL);
  assert(key->name != NULL);

  bzero(key->name, key->nleng);
  bzero(key->data, key->dleng);
  
  memcpy(key->name, name, key->nleng);
  memcpy(key->data, data, key->dleng);
}

int
hash_key_CompareToValue (key, name, nleng)
  hash_key_t * key;
  unsigned const char * name;
  unsigned long nleng;
{
  int pos = 0;
  int max = 0;
  unsigned const char * ptr1;
  unsigned const char * ptr2;
  
  assert(key   != NULL);
  assert(name  != NULL);
  assert(nleng >= 1);

  // before actually comparing the data, byte for byte, just check
  // the stated lengths of the two values, and return if they do
  // not match; 1 if key is greater; -1 if data is greater.
  if (key->nleng != nleng)
    return key->nleng > nleng ? 1 : -1;

  ptr1 = key->name;
  ptr2 = name;
  max  = key->nleng > nleng ? key->nleng : nleng;
  pos  = 0;

  while (pos < max)
  {
    if (*ptr1 != *ptr2)
      return *ptr1 > *ptr2 ? 1 : -1;
    
    *ptr1++;
    *ptr2++;
    pos++;
  }
  
  return 0;
}

int
hash_key_Compare (key1, key2)
  hash_key_t * key1;
  hash_key_t * key2;
{
  int pos = 0;
  int max = 0;
  unsigned const char * ptr1;
  unsigned const char * ptr2;
  
  assert(key1 != NULL);
  assert(key2 != NULL);
  
  // before actually comparing the data, byte for byte, just check
  // the stated lengths of the two values, and return if they do
  // not match; 1 if key is greater; -1 if data is greater.
  if (key1->nleng != key2->nleng)
    return key1->nleng > key2->nleng ? 1 : -1;
  
  ptr1 = key1->name;
  ptr2 = key2->name;
  max  = key1->nleng > key2->nleng ? key1->nleng : key2->nleng;
  pos  = 0;
  
  while (pos < max)
  {
    if (*ptr1 != *ptr2)
      return *ptr1 > *ptr2 ? 1 : -1;
    
    *ptr1++;
    *ptr2++;
    pos++;
  }
  
  return 0;
}

void
hash_key_SetValue (key, data, dleng)
  hash_key_t * key;
  unsigned const char * data;
  unsigned long dleng;
{
  assert(key  != NULL);
  assert(data != NULL);
  
  key->dleng = dleng;
  
  if (key->data == NULL)
    key->data = (unsigned char *)malloc(key->dleng + 1);
  else
    key->data = (unsigned char *)realloc(key->data, key->dleng + 1);
  
  assert(key->data != NULL);
  
  bzero(key->data, key->dleng + 1);
  memcpy(key->data, data, key->dleng);
}

void
hash_key_Destroy (key)
  hash_key_t * key;
{
  assert(key != NULL);
  
  if (key->name != NULL)
    free(key->name);
  
  if (key->data != NULL)
    free(key->data);

  key->name = NULL;
  key->data = NULL;
  
  key->nleng = 0;
  key->dleng = 0;
  
  free(key);
}
