#include <assert.h>           // assert()
#include <stdio.h>            // fopen()
#include <stdlib.h>           // malloc()
#include <strings.h>          // bzero()

#include "mpeg.h"

static const unsigned int mpeg_style[][4] =
{
  {0, 5, 4, 3},     // mpeg-2
  {0, 2, 1, 0}      // mpeg-1
};

static const unsigned int mpeg_frequency[][2] =
{
  {22050, 44100},
  {24000, 48000},
  {16000, 32000},
  {    0,     0}
};

//
// the numbers above each column represent the
// mpeg type and layer, respective. the numbers
// to the right of each row represent the value
// encoded in the mpeg file.
//
static const unsigned short mpeg_bitrate[][6] =
{
// 1,1    1,2     1,3     2,1     2,2     2,3
  {  0,     0,      0,      0,      0,      0},         // 0000
  { 32,    32,     32,     32,     32,      8},         // 0001
  { 64,    48,     40,     64,     48,     16},         // 0010
  { 96,    56,     48,     96,     56,     24},         // 0011
  {128,    64,     56,    128,     64,     32},         // 0100
  {160,    80,     64,    160,     80,     64},         // 0101
  {192,    96,     80,    192,     96,     80},         // 0110
  {224,   112,     96,    224,    112,     56},         // 0111
  {256,   128,    112,    256,    128,     64},         // 1000
  {288,   160,    128,    288,    160,    128},         // 1001
  {320,   192,    160,    320,    192,    160},         // 1010
  {352,   223,    192,    352,    224,    112},         // 1011
  {384,   256,    224,    384,    256,    128},         // 1100
  {416,   320,    256,    416,    320,    256},         // 1101
  {448,   384,    320,    448,    384,    320}          // 1110
};

int
mpeg_Load (song, path, spos)
  daap_song * song;
  char      * path;
  int         spos;
{
  FILE * file;
  
  // all of these values are used over and over again
  // for each frame in this file. they're just here,
  // instead of in the loop itself, because I need to
  // reference them after the loop.
  //
  unsigned int x_bitrate;             // 4 bit bitrate code
  unsigned int x_layer;               // 2 bit layer code
  unsigned int x_mpegid;              // 1 bit mpeg id
  unsigned int x_frequ;               // 2 bit frequency code
  unsigned int x_padding;             // 1 bit padding code
  unsigned int x_protect;             // 1 bit protection code
  unsigned int z_bitrate;             // bit rate of the current frame
  unsigned int z_frequ;               // sample rate of the current frame
  unsigned int z_framebytes;          // byte length of the current frame
  
  unsigned char data[1000];           // thousand byte file read buffer
  unsigned char head[4];              // used for the four byte frame header
  
  double ravg = 0.0;                  // bitrate average
  double rlow = 0.0;                  // bitrate low
  double rhi  = 0.0;                  // bitrate hi
  int    rcnt = 0;                    // frame count
  
  unsigned long fpos = 0;             // our current position in the file
  size_t rleng = 0;                   // length of data read on last fread()
  int i = 0;                          // our current position in the buffer
  unsigned int playtime = 0;          // play time in milliseconds

  assert(song != NULL);
  assert(path != NULL);
  
  fprintf(stderr, "mpeg_Load.. %s: LOADING @ %d\n", path, spos);
  
  // open the file, and die if we fail
  file = fopen(path, "r");
  assert(file != NULL);

  // zero out all of the arrays for good luck
  bzero(&head,  4);
  bzero(&data,  1000);

  // in order to avoid problems in identifying what is, and is not
  // actually an mpeg frame header, I have added a seek-to-this-pos-
  // ition-to-start-reading-from feature, which is set by the previous
  // call to the id3_Load() function. so, basically, we start reading
  // where the ID3 parser left off.
  if (spos > 0)
    fseek(file, spos, SEEK_SET);
  
  // initialize the buffer with the first chunk of the file
  rleng = fread(&data, 1, 1000, file);

  // scan through the mpeg audio file until we find the
  // first frame header. we exit the loop at that point
  while (1)
  {
    int found = 0;

    // when we find two bytes that match the description of a
    // frame header, then break out of these two loops.
    for (i = 0; i < rleng - 3; i++, fpos++) {
      if (data[i] == 0xFF && (data[i+1] & 0xF0) == 0xF0) {
        memcpy(&head, data + i, 4);
        found = 1;
        break;
      }
    }

    // if it seems that we have found the first mpeg frame
    // header in the file, then break out of this loop
    if (found == 1)
      break;

    // if the amount of data that we read on the last call
    // to fread() was less than the amount of data that we
    // tried to read, then we are most likely at the end of
    // the file, and should utterly give up at this point.
    if (rleng < 1000) {
      fprintf(stderr, "mpeg_Load.. %s: only read %d bytes, exiting\n", path, (int)rleng);
      break;
    }

    // move the last three bytes of the read buffer down up
    // to the front of the buffer, and fill in the remaining
    // space with data from the file. those three bytes represent
    // what could possibly be the first three bytes of a frame
    // header, and since I don't want to worry about fragmentation
    // like that, I just account for it in this manner.
    memcpy(&data, data + 997, 3);
    rleng = fread(data + 3, 1, 997, file) + 3;
  }

  // scan through the rest of the file, starting at our current
  // position, which should be the first frame header, and skip
  // through the rest of the file one frame at a time, calculating
  // the distance to the next frame by looking at each frame as
  // we go along.
  //
  // we also update various values, such as the hi, low, and average
  // bit rates, the frame count, and other useful bits of data.
  //
  while (1)
  {
    // this is repetitive the first time through the loop ... it
    // just copies the frame header into our little four byte array
    memcpy(&head, data + i, 4);

    // if our mpeg audio frame header doesn't look the least bit like
    // an mpeg audio frame header, and lacking other senses with which
    // to analyze it (smell, touch, etc), we should assume that what
    // we have here is in fact _not_ the header we're looking for, and
    // exit this faitfully doomed loop in search of better things to do.
    if (head[0] != 0xFF || (head[1] & 0xF0) != 0xF0)
      break;
    
    // parse all of the flags we're interested in, out of
    // the frame header
    x_bitrate = (head[2] >> 4) & 0x0F;     // 4 bit bitrate code
    x_layer   = (head[1] >> 1) & 0x01;     // 2 bit layer code
    x_mpegid  = (head[1] >> 7) & 0x01;     // 1 bit mpeg id
    x_frequ   = (head[2] >> 2) & 0x03;     // 2 bit frequency code
    x_padding = (head[2] >> 1) & 0x01;     // 1 bit padding code
    x_protect = (head[1] >> 4) & 0x01;     // 1 bit protection code

    // convert the mpeg codes into the actual values we care about
    z_bitrate = mpeg_bitrate   [x_bitrate][mpeg_style[x_mpegid][x_layer]];
    z_frequ   = mpeg_frequency [x_frequ][x_mpegid];

    // calculate the length (in bytes) of this frame based on its bit rate,
    // and the frequency ... plus whether there is any padding after the frame.
    if (x_layer == MPEG_LAYER_2 || x_layer == MPEG_LAYER_3)
      z_framebytes = ((144 * (z_bitrate * 1000)) / z_frequ) + (x_padding ? 1 : 0);
    else if (x_layer == MPEG_LAYER_1)
      z_framebytes = (((12 * (z_bitrate * 1000)) / z_frequ) + (x_padding ? 4 : 0)) * 4;
   
    // update a running average of the bitrate over all of the frames
    if (ravg != z_bitrate)
      ravg = ((ravg * rcnt) + z_bitrate) / (rcnt + 1);

    // track the bit rate hi and low "marks" over the whole file
    if (rlow == 0 || rlow > z_bitrate)
      rlow = z_bitrate;
    if (z_bitrate > rhi)
      rhi  = z_bitrate;

    // the total number of frames encountered thus far
    rcnt++;

    // if we have calculated that the size of the frame is less than or
    // equal to zero, then we have run into a problem. the only place I
    // have actually encountered this is when I mistake an image contained
    // in an ID3 tag for a frame header. I have since then taken steps to
    // prevent that from happening, but, on the off-chance that it does,
    // this should prevent it.
    if (z_framebytes <= 0) {
      fprintf(stderr, "mpeg_Load.. %s: invalid frame length, %d\n", path, z_framebytes);
      fclose(file);
      return 0;
    }
    
    // fpos maintains our position in the file. we are now
    // skipping that forward by the length of the frame we
    // are now processing
    fpos += z_framebytes;

    // we are using a 1000 byte internal buffer for this file
    // scan. as long as the buffer holds all of the data we are
    // interested in, then we just scan through the buffer. when
    // it becomes inadequate, we fseek() through the file to the
    // start point of the next frame, and write the next 1000
    // bytes into the buffer (a few lines down)
    if (i + z_framebytes < rleng - 3) {
      i += z_framebytes;
      continue;
    }

    // we always try to read 1000 bytes at a time, and we only reach
    // this point if all of the data read thus far is inadequate to
    // process the next frame. so, if we read less than 1000 bytes
    // on the last read, then we are at the end of the file. stop.
    if (rleng < 1000)
      break;

    // seek forward to the point of the header for the next frame, and
    // re-fill the buffer starting at that point.
    fseek(file, z_framebytes - (1000 - i), SEEK_CUR);
    rleng = fread(data, 1, 1000, file);
    i = 0;
  }

  // if this is a layer II or III file, then there should be exactly
  // 1152 samples in each frame. for layer I files, there are 384.
  // if we divide the number of samples by the frequency (eg, 44100)
  // and multiply it by the number of frames containing that number of
  // samples, and multiply that by 1000, then we get something that
  // should suffice in accuracy for showing the play length in seconds.
  if (x_layer == MPEG_LAYER_2 || x_layer == MPEG_LAYER_3)
    playtime = ((1152.0 / z_frequ) * rcnt) * 1000;
  else if (x_layer == MPEG_LAYER_1)
    playtime = (( 384.0 / z_frequ) * rcnt) * 1000;
  
  fprintf(stderr, "            bitrate_hi=%f; bitrate_low=%f; bitrate_avg=%f, total-frames=%05d\n",
          rhi, rlow, ravg, rcnt);
  fprintf(stderr, "            bitrate-code=%d, layer-code=%d, mpegid=%d, frequencyid=%d\n",
          x_bitrate, x_layer, x_mpegid, x_frequ);
  fprintf(stderr, "            padding=%d, protected=%d, play-time=%u, sample-rate=%d\n",
          x_padding, x_protect, playtime, z_frequ);

  // apply this hard-earned knowledge regarding this file to the song
  // structure so that it can be sent back to iTunes
  asbr_Init(&song->asbr, (daap_int)ravg);
  assr_Init(&song->assr, (daap_int)z_frequ);
  astm_Init(&song->astm, (daap_int)playtime);
  
  fclose(file);
  
  return 0;
}
