#include <stdio.h>          // For printf()
#include <stdlib.h>         // For exit() etc.
#include <string.h>         // For strlen() etc.
#include <unistd.h>         // For select()
#include <errno.h>          // For errno, EINTR
#include <arpa/inet.h>      // For inet_addr()
#include <netinet/in.h>     // For INADDR_NONE
#include <netdb.h>          // For gethostbyname()
#include <assert.h>         // For assert()
#include <signal.h>         // For SIG stuff
#include <sys/time.h>       // For struct timeval
#include <pthread.h>        // For pthreads, duh
#include <sys/uio.h>        // For read() etc
#include <sys/types.h>      // For socket stuff
#include <sys/socket.h>     // For socket stuff
#include <strings.h>        // For bzero()

#include "mDNSClientAPI.h"  // Defines the interface to the client layer above
#include "mDNSPosix.h"      // Defines the specific types needed to run mDNS on this platform
#include "iCast.h"          // Defines struct Conn, etc

//
// The following HTTP request is an example of what iTunes sends when
// a shared music library is clicked upon.
//
//   GET /server-info HTTP/1.1
//   Host: 10.0.1.50:5090
//   Accept: */*
//   User-Agent: iTunes/4.2 (Macintosh; N; PPC)
//   Client-DAAP-Version: 2.0
//   Client-DAAP-Access-Index: 1
//   Accept-Encoding: gzip
//
// And, although it's not entirely relevant, iTunes listens on port
// 3689 for such requests. However, it should not be necessary to hard
// code this, as such information is advertised in the mDNS packets.
//
// File list:
//
//   server-info
//   content-codes
//   login
//   update [?[session-id=][revision-number=1]]
//   databases [?[session-id=][revision-number=3]]
//   databases/33/items [?[type=music][meta=][session-id=][revision-number=3]]
//   databases/33/containers/9945/items [?[type=music][meta=][session-id=][revision-number=3]]
//



#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Globals
#endif

static mDNS mDNSStorage;
static mDNS_PlatformSupport PlatformStorage;

static const char *gProgramName = "iCast";

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Signals
#endif

static volatile mDNSBool gHault;
static volatile mDNSBool gReceivedSigHup;
static volatile mDNSBool gReceivedSigUsr1;
static volatile mDNSBool gReceivedSigUsr2;

static void
HandleSigUsr1 (int sigraised)
{
  assert(sigraised == SIGUSR1);
  gReceivedSigUsr1 = mDNStrue;
}

static void
HandleSigUsr2 (int sigraised)
{
  assert(sigraised == SIGUSR2);
  gReceivedSigUsr2 = mDNStrue;
}

static void
HandleSigHup (int sigraised)
{
  assert(sigraised = SIGHUP);
  gReceivedSigHup = mDNStrue;
}

static void
HandleSigInt (int sigraised)
{
  assert(sigraised == SIGINT);
  gHault = mDNStrue;
}

static void
HandleSigQuit (int sigraised)
{
  assert(sigraised == SIGQUIT);
  mDNS_Close(&mDNSStorage);
  exit(0);
}

static const char kDefaultPIDFile[]       = "/var/run/iCast.pid";
static const char kDefaultServiceType[]   = "_daap._tcp.";
static const char kDefaultServiceDomain[] = "local.";
static const char kDefaultServiceRoot[]   = "./files";
static const char kDefaultServicePass[]   = "corvette";

enum {
  kDefaultPortNumber = 5090
};

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** DAAP Stuff
#endif

static inline int
WriteDAAPLong (char * data, char * name, unsigned long long valu)
{
  unsigned int leng = htonl(8);

  unsigned int valu1 = 0;
  unsigned int valu2 = 0;
  
  valu1 = htonl((valu      ) & 0xFFFFFFFF);
  valu2 = htonl((valu >> 32) & 0xFFFFFFFF);
  
  memcpy(data +  0, name,   4);
  memcpy(data +  4, &leng,  4);
  memcpy(data +  8, &valu2, 4);
  memcpy(data + 12, &valu1, 4);
  
  return 16;
}

static inline int
WriteDAAPInt (char * data, char * name, unsigned int valu)
{
  unsigned int leng = htonl(4);

  valu = htonl(valu);
  
  memcpy(data + 0, name,  4);
  memcpy(data + 4, &leng, 4);
  memcpy(data + 8, &valu, 4);
  
  return 12;
}

static inline int
WriteDAAPShort (char * data, char * name, unsigned short valu)
{
  unsigned int leng = htonl(2);
  
  valu = htonl(valu);
  
  memcpy(data + 0, name,  4);
  memcpy(data + 4, &leng, 4);
  memcpy(data + 8, &valu, 2);
  
  return 10;
}

static inline int
WriteDAAPByte (char * data, char * name, unsigned char valu)
{
  unsigned int leng = htonl(1);
  
  memcpy(data + 0, name,  4);
  memcpy(data + 4, &leng, 4);
  memcpy(data + 8, &valu, 1);
  
  return 9;
}

static inline int
WriteDAAPVersion (char * data, char * name, unsigned short major, short minor)
{
  unsigned int leng = 4;
  
  unsigned short maj = htons(major);
  unsigned short min = htons(minor);
  
  memcpy(data +  0, name,   4);
  memcpy(data +  4, &leng,  4);
  memcpy(data +  8, &maj,   2);
  memcpy(data + 10, &min,   2);
  
  return 12;
}

static inline int
WriteDAAPList (char * data, char * name, char * valu, unsigned int leng)
{
  leng = htonl(leng);
  
  memcpy(data + 0, name,  4);
  memcpy(data + 4, &leng, 4);
  memcpy(data + 8, valu,  ntohl(leng));
  
  return 8 + ntohl(leng);
}

static inline int
WriteDAAPString (char * data, char * name, char * valu, unsigned int leng)
{
  if (valu == NULL && leng != 0)
    fprintf(stderr, "          ERROR: leng [%d] != 0\n", (int)leng);
  else if (valu != NULL && leng != strlen(valu))
    fprintf(stderr, "          ERROR: leng [%d] != strlen() [%d]\n", (int)leng, (int)strlen(valu));
  
  leng = htonl(leng);
  
  memcpy(data + 0, name,  4);
  memcpy(data + 4, &leng, 4);
  memcpy(data + 8, valu,  ntohl(leng));
  
  return 8 + ntohl(leng);
}

static inline int
WriteDAAPNameAndLength (char * data, char * name, int leng)
{
  leng = htonl(leng);
  
  memcpy(data + 0, name,  4);
  memcpy(data + 4, &leng, 4);
  
  return 8;
}

static inline int
WriteDAAPMdcl (char * data, char * numb, char * name, unsigned short type)
{
  int datalen = 0;
  
  datalen += WriteDAAPList    (data + datalen, "mdcl", NULL,  0);   // dmap.dictionary              dictionary entry for each content code
  datalen += WriteDAAPString  (data + datalen, "mcnm", numb,  4);   // dmap.contentcodesnumber      item number
  datalen += WriteDAAPString  (data + datalen, "mcna", name,  strlen(name));   // dmap.contentcodesname        item name
  datalen += WriteDAAPShort   (data + datalen, "mcty", type    );   // dmap.contentcodestype        item type code
  
  WriteDAAPNameAndLength(data, "mdcl", datalen - 8);
  
  return datalen;
}

static void
WriteDAAPFile (char * data, int leng, char * name)
{
  FILE * file;
  
  if (NULL == (file = fopen(name, "w"))) {
    fprintf(stderr, "WriteDAAPFile.. fopen() failed, '%s'\n", name);
    return;
  }
  
  fwrite(data, sizeof(char), leng, file);
  fclose(file);
}

static char *
CreateServerInfo (int * dleng)
{
  char * data;
  int datalen;
  int datamax;
  
  if (NULL == (data = (char *)malloc(sizeof(char) * 200))) {
    fprintf(stderr, "CreateServerInfo.. malloc() failed [1]\n");
    exit(1);
  }
  
  datalen = 0;
  datamax = 200;

  bzero(data, datamax);

  datalen += WriteDAAPList    (data + datalen, "msrv", NULL, 0);    // dmap.serviceinforesponse     server-info response data list
  datalen += WriteDAAPInt     (data + datalen, "mstt", 200);        // dmap.status                  status code (200)
  datalen += WriteDAAPVersion (data + datalen, "mpro", 2, 0);       // dmap.protocolversion         dmap protocol version
  datalen += WriteDAAPVersion (data + datalen, "apro", 2, 0);       // daap.protocolversion         daap protocol
  datalen += WriteDAAPString  (data + datalen, "minm", "iCast", 5); // dmap.itemname                server name
  datalen += WriteDAAPByte    (data + datalen, "mslr", 0);          // dmap.loginrequired           is login required?
  datalen += WriteDAAPInt     (data + datalen, "mstm", 1800);       // dmap.timeoutinterval         timeout interval
  datalen += WriteDAAPByte    (data + datalen, "msal", 1);          // dmap.supportsautologout      does the server support auto-logout?
  datalen += WriteDAAPByte    (data + datalen, "msup", 1);          // dmap.supportsupdate          does the server support update?
  datalen += WriteDAAPByte    (data + datalen, "mspi", 1);          // dmap.supportspersistentids   does the server support persistent ids?
  datalen += WriteDAAPByte    (data + datalen, "msex", 1);          // dmap.supportsextensions      does the server support extensions?
  datalen += WriteDAAPByte    (data + datalen, "msbr", 1);          // dmap.supportsbrowse          does the server support browsing?
  datalen += WriteDAAPByte    (data + datalen, "msqy", 1);          // dmap.supportsquery           does the server support queries?
  datalen += WriteDAAPByte    (data + datalen, "msix", 1);          // dmap.supportsindex           does the server support indexing?
  datalen += WriteDAAPByte    (data + datalen, "msrs", 1);          // dmap.supportsresolve         does the server support resolve?
  datalen += WriteDAAPInt     (data + datalen, "msdc", 1);          // dmap.databasescount          <no description available>

  WriteDAAPNameAndLength(data, "msrv", datalen - 8);
  
  WriteDAAPFile(data, datalen, "./01_server-info.out");

  *dleng = datalen;
  
  return data;
}

static char *
CreateContentCodes (int * dleng)
{
  char * data;
  int datalen;
  int datamax;
  
  if (NULL == (data = (char *)malloc(sizeof(char) * 10000))) {
    fprintf(stderr, "CreateContentCodes.. malloc() failed [1]\n");
    exit(1);
  }
  
  datalen = 0;
  datamax = 5000;
  
  bzero(data, datamax);
  
  datalen += WriteDAAPList    (data + datalen, "mccr", NULL, 0);    // dmap.contentcodesresponse    response to content-codes request
  datalen += WriteDAAPInt     (data + datalen, "mstt", 200);        // dmap.status                  status code (200)
  
  datalen += WriteDAAPMdcl(data + datalen, "miid", "dmap.itemid",                       DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "minm", "dmap.itemname",                     DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "mikd", "dmap.itemkind",                     DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "mper", "dmap.persistentid",                 DAAP_TYPE_LONG);
  datalen += WriteDAAPMdcl(data + datalen, "mcon", "dmap.container",                    DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "mcti", "dmap.containeritemid",              DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "mpco", "dmap.parentcontainerid",            DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "mstt", "dmap.status",                       DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "msts", "dmap.statusstring",                 DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "mimc", "dmap.itemcount",                    DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "mctc", "dmap.containercount",               DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "mrco", "dmap.returnedcount",                DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "mtco", "dmap.specifiedtotalcount",          DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "mlcl", "dmap.listing",                      DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "mlit", "dmap.listingitem",                  DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "mbcl", "dmap.bag",                          DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "mdcl", "dmap.dictionary",                   DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "msrv", "dmap.serverinforesponse",           DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "msau", "dmap.authenticationmethod",         DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "mslr", "dmap.loginrequired",                DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "mpro", "dmap.protocolversion",              DAAP_TYPE_VERSION);
  datalen += WriteDAAPMdcl(data + datalen, "msal", "dmap.supportsautologout",           DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "msup", "dmap.supportsupdate",               DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "mspi", "dmap.supportspersistentids",        DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "msex", "dmap.supportsextensions",           DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "msbr", "dmap.supportsbrowse",               DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "msqy", "dmap.supportsquery",                DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "msix", "dmap.supportsindex",                DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "msrs", "dmap.supportsresolve",              DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "mstm", "dmap.timeoutinterval",              DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "msdc", "dmap.databasescount",               DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "mlog", "dmap.loginresponse",                DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "mlid", "dmap.sessionid",                    DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "mupd", "dmap.updateresponse",               DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "musr", "dmap.serverrevision",               DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "muty", "dmap.updatetype",                   DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "mudl", "dmap.deletedidlisting",             DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "mccr", "dmap.contentcodesresponse",         DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "mcnm", "dmap.contentcodesnumber",           DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "mcna", "dmap.contentcodesname",             DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "mcty", "dmap.contentcodestype",             DAAP_TYPE_SHORT);
  datalen += WriteDAAPMdcl(data + datalen, "apro", "daap.protocolversion",              DAAP_TYPE_VERSION);
  datalen += WriteDAAPMdcl(data + datalen, "avdb", "daap.serverdatabases",              DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "abro", "daap.databasebrowse",               DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "abal", "daap.browsealbumlisting",           DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "abar", "daap.browseartistlisting",          DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "abcp", "daap.browsecomposerlisting",        DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "abgn", "daap.browsegenrelisting",           DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "adbs", "daap.databasesongs",                DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "asal", "daap.songalbum",                    DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "agrp", "daap.songgrouping",                 DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "asar", "daap.songartist",                   DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "asbt", "daap.songbeatsperminute",           DAAP_TYPE_SHORT);
  datalen += WriteDAAPMdcl(data + datalen, "asbr", "daap.songbitrate",                  DAAP_TYPE_SHORT);
  datalen += WriteDAAPMdcl(data + datalen, "ascm", "daap.songcomment",                  DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "asco", "daap.songcompilation",              DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "ascp", "daap.songcomposer",                 DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "asda", "daap.songdateadded",                DAAP_TYPE_DATE);
  datalen += WriteDAAPMdcl(data + datalen, "asdm", "daap.songdatemodified",             DAAP_TYPE_DATE);
  datalen += WriteDAAPMdcl(data + datalen, "asdc", "daap.songdisccount",                DAAP_TYPE_SHORT);
  datalen += WriteDAAPMdcl(data + datalen, "asdn", "daap.songdiscnumber",               DAAP_TYPE_SHORT);
  datalen += WriteDAAPMdcl(data + datalen, "asdb", "daap.songdisabled",                 DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "aseq", "daap.songeqpreset",                 DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "asfm", "daap.songformat",                   DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "asgn", "daap.songgenre",                    DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "asdt", "daap.songdescription",              DAAP_TYPE_STRING);
//datalen += WriteDAAPMdcl(data + datalen, "asrv", "daap.songrelativevolume",           0x0002);
  datalen += WriteDAAPMdcl(data + datalen, "assr", "daap.songsamplerate",               DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "assz", "daap.songsize",                     DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "asst", "daap.songstarttime",                DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "assp", "daap.songstoptime",                 DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "astm", "daap.songtime",                     DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "astc", "daap.songtrackcount",               DAAP_TYPE_SHORT);
  datalen += WriteDAAPMdcl(data + datalen, "astn", "daap.songtracknumber",              DAAP_TYPE_SHORT);
  datalen += WriteDAAPMdcl(data + datalen, "asur", "daap.songuserrating",               DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "asyr", "daap.songyear",                     DAAP_TYPE_SHORT);
  datalen += WriteDAAPMdcl(data + datalen, "asdk", "daap.songdatakind",                 DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "asul", "daap.songdataurl",                  DAAP_TYPE_STRING);
  datalen += WriteDAAPMdcl(data + datalen, "aply", "daap.databaseplaylists",            DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "abpl", "daap.baseplaylist",                 DAAP_TYPE_BYTE);
  datalen += WriteDAAPMdcl(data + datalen, "apso", "daap.playlistsongs",                DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "arsv", "daap.resolve",                      DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "arif", "daap.resolveinfo",                  DAAP_TYPE_LIST);
  datalen += WriteDAAPMdcl(data + datalen, "aeNV", "com.apple.itunes.norm-volume",      DAAP_TYPE_INT);
  datalen += WriteDAAPMdcl(data + datalen, "aeSP", "com.apple.itunes.smart-playlist",   DAAP_TYPE_BYTE);
  
  WriteDAAPNameAndLength(data, "mccr", datalen - 8);
  
  WriteDAAPFile(data, datalen, "./02_content-codes.out");
  
  *dleng = datalen;
  
  return data;
}

static char *
CreateLogin (int * dleng)
{
  char * data;
  int datalen;
  int datamax;
  
  if (NULL == (data = (char *)malloc(sizeof(char) * 200))) {
    fprintf(stderr, "CreateLogin.. malloc() failed [1]\n");
    exit(1);
  }
  
  datalen = 0;
  datamax = 200;
  
  bzero(data, datamax);

  datalen += WriteDAAPList    (data + datalen, "mlog", NULL, 0);    // dmap.loginresponse           response to a login request
  datalen += WriteDAAPInt     (data + datalen, "mstt", 200);        // dmap.status                  status code (200)
  datalen += WriteDAAPInt     (data + datalen, "mlid", 0x1234);     // dmap.sessionid               the session id for the login session
  
  WriteDAAPNameAndLength(data, "mlog", datalen - 8);
  
  WriteDAAPFile(data, datalen, "./03_login.out");
  
  *dleng = datalen;
  
  return data;
}

static char *
CreateUpdate (int * dleng)
{
  char * data;
  int datalen;
  int datamax;
  
  if (NULL == (data = (char *)malloc(sizeof(char) * 200))) {
    fprintf(stderr, "CreateUpdate.. malloc() failed [1]\n");
    exit(1);
  }
  
  datalen = 0;
  datamax = 200;
  
  bzero(data, datamax);
  
  datalen += WriteDAAPList    (data + datalen, "mupd", NULL, 0);    // dmap.updateresponse          response to an update request
  datalen += WriteDAAPInt     (data + datalen, "mstt", 200);        // dmap.status                  status code (200)
  datalen += WriteDAAPInt     (data + datalen, "musr", 3);          // dmap.serverrevision          revision to use for requests
  
  WriteDAAPNameAndLength(data, "mupd", datalen - 8);
  
  WriteDAAPFile(data, datalen, "./04_update.out");
  
  *dleng = datalen;
  
  return data;
}

static char *
CreateDatabasesList (int * dleng)
{
  char * data;
  int datalen;
  int datamax;
  
  if (NULL == (data = (char *)malloc(sizeof(char) * 200))) {
    fprintf(stderr, "CreateDatabases.. malloc() failed [1]\n");
    exit(1);
  }
  
  datalen = 0;
  datamax = 200;
  
  bzero(data, datamax);
  
  datalen += WriteDAAPList    (data + datalen, "avdb", NULL, 0);    // daap.serverdatabases         response to /databases
  datalen += WriteDAAPInt     (data + datalen, "mstt", 200);        // dmap.status                  status code (200)
  datalen += WriteDAAPByte    (data + datalen, "muty", 0);          // dmap.updatetype              <no description available>
  datalen += WriteDAAPInt     (data + datalen, "mtco", 1);          // dmap.specifiedtotalcount     number of items in response to a request
  datalen += WriteDAAPInt     (data + datalen, "mrco", 1);          // dmap.returnedcount           number of items returned in a request
  datalen += WriteDAAPList    (data + datalen, "mlcl", NULL, 0);    // dmap.listing                 a list
  datalen += WriteDAAPList    (data + datalen, "mlit", NULL, 0);    // dmap.listingitem             a single item in said list
  datalen += WriteDAAPInt     (data + datalen, "miid", 0x20);       // dmap.itemid                  database id
  datalen += WriteDAAPLong    (data + datalen, "mper", 0x73a30788); // dmap.persistentid            database persistent id
  datalen += WriteDAAPString  (data + datalen, "minm", "iCast Music", 11);  // dmap.itemname        databasename
  datalen += WriteDAAPInt     (data + datalen, "mimc", 1);          // dmap.itemcount               number of songs in database
  datalen += WriteDAAPInt     (data + datalen, "mctc", 1);          // dmap.containercount          number of playlists in database
  
  WriteDAAPNameAndLength(data, "avdb", datalen - 8);
  WriteDAAPNameAndLength(data + 53, "mlcl", datalen - (53 + 8));
  WriteDAAPNameAndLength(data + 61, "mlit", datalen - (61 + 8));
  
  WriteDAAPFile(data, datalen, "./05_databases.out");
  
  *dleng = datalen;
  
  return data;
}

static char *
CreateListSongs (int * dleng)
{
  char * data;
  int datalen;
  int datamax;
  
  if (NULL == (data = (char *)malloc(sizeof(char) * 500))) {
    fprintf(stderr, "CreateGetADatabase.. malloc() failed [1]\n");
    exit(1);
  }
  
  datalen = 0;
  datamax = 200;
  
  bzero(data, datamax);
  
  datalen += WriteDAAPList    (data + datalen, "adbs", NULL, 0);        // daap.playlistsongs         response to /databases/<id>/items
  datalen += WriteDAAPInt     (data + datalen, "mstt", 200);            // dmap.status                status code (200)
  datalen += WriteDAAPByte    (data + datalen, "muty", 0);              // dmap.updatetype            <no description available>
  datalen += WriteDAAPInt     (data + datalen, "mtco", 1);              // dmap.specifiedtotalcount   number of items in response to a request
  datalen += WriteDAAPInt     (data + datalen, "mrco", 1);              // dmap.returnedcount         number of items returned in a request
  datalen += WriteDAAPList    (data + datalen, "mlcl", NULL, 0);        // dmap.listing               a list
  datalen += WriteDAAPList    (data + datalen, "mlit", NULL, 0);        // dmap.listingitem           a single item in said list
  
  datalen += WriteDAAPByte    (data + datalen, "mikd", 0x02);           // dmap.itemkind              only '2' is used; for audio?
  datalen += WriteDAAPInt     (data + datalen, "miid", 0x01);           // dmap.itemid                a unique id for each item
  datalen += WriteDAAPString  (data + datalen, "minm", "Song.mp3", 8);  // dmap.itemname              item name (eg, the track name)
  datalen += WriteDAAPLong    (data + datalen, "mper", 0x1234LL);       // dmap.persistentid          a persistent id
  datalen += WriteDAAPString  (data + datalen, "asal", "Album", 5);     // daap.songalbum             album
  datalen += WriteDAAPString  (data + datalen, "asar", "Artist", 6);    // daap.songartist            artist
  datalen += WriteDAAPShort   (data + datalen, "asbr", 128);            // daap.songbitrate           bitrate
  datalen += WriteDAAPShort   (data + datalen, "asbt", 0);              // daap.songbeatsperminute    beats per minute
  datalen += WriteDAAPString  (data + datalen, "ascm", "Comment", 7);   // daap.songcomment           comment
  datalen += WriteDAAPByte    (data + datalen, "asco", 0x00);           // daap.songcompilation       compilation
//datalen += WriteDAAPString  (data + datalen, "ascp", "Composer", 8);  // daap.songcomposer          composer
//datalen += WriteDAAPInt     (data + datalen, "asda", 0x3e95b443);     // daap.songdateadded         date added
//datalen += WriteDAAPInt     (data + datalen, "asdm", 0x3e95b443);     // daap.songdatemodified      date modified
  datalen += WriteDAAPShort   (data + datalen, "asdc", 0x01);           // daap.songdisccount         disc count
  datalen += WriteDAAPShort   (data + datalen, "asdn", 0x01);           // daap.songdiscnumber        disc number
  datalen += WriteDAAPByte    (data + datalen, "asdb", 0x00);           // daap.songdisabled          disabled
  datalen += WriteDAAPString  (data + datalen, "aseq", NULL, 0);        // daap.songeqpreset          eq preset
  datalen += WriteDAAPString  (data + datalen, "asfm", "mp3", 3);       // daap.songformat            format
  datalen += WriteDAAPString  (data + datalen, "asgn", "Genre", 5);     // daap.songgenre             genre
  datalen += WriteDAAPString  (data + datalen, "asdt", "Desc", 4);      // daap.songdescription       description
//datalen += WriteDAAPByte    (data + datalen, "asrv", 0x00);           // daap.songrelativevolume    relative volume
  datalen += WriteDAAPInt     (data + datalen, "assr", 0xAC44);         // daap.songsamplerate        sample rate
//datalen += WriteDAAPInt     (data + datalen, "assz", 0x6E012D);       // daap.songsize              size
  datalen += WriteDAAPInt     (data + datalen, "assz", 0x5E528E);       // daap.songsize              size
//datalen += WriteDAAPInt     (data + datalen, "asst", 0x00);           // daap.songstarttime         start time    (in milliseconds)
//datalen += WriteDAAPInt     (data + datalen, "assp", 0x00);           // daap.songstoptime          stop time     (in milliseconds)
//datalen += WriteDAAPInt     (data + datalen, "astm", 0x03851F);       // daap.songtime              time          (in milliseconds)
  datalen += WriteDAAPShort   (data + datalen, "astc", 0x0001);         // daap.songtrackcount        track count
  datalen += WriteDAAPShort   (data + datalen, "astn", 0x0001);         // daap.songtracknumber       track number
  datalen += WriteDAAPByte    (data + datalen, "asur", 0x00);           // daap.songuserrating        user rating
  datalen += WriteDAAPShort   (data + datalen, "asyr", 2000);           // daap.songyear              year
  datalen += WriteDAAPByte    (data + datalen, "asdk", 0x00);           // daap.songdatakind          data kind
  datalen += WriteDAAPString  (data + datalen, "asul", NULL, 0);        // daap.songdataurl           data url
  datalen += WriteDAAPInt     (data + datalen, "aeNV", 0x00);           // com.apple.itunes.norm-volume
  datalen += WriteDAAPByte    (data + datalen, "aeSP", 0x00);           // com.apple.itunes.smart-playlist
  
  WriteDAAPNameAndLength(data,      "adbs", datalen - 8);
  WriteDAAPNameAndLength(data + 53, "mlcl", datalen - (53 + 8));
  WriteDAAPNameAndLength(data + 61, "mlit", datalen - (61 + 8));
  
  WriteDAAPFile(data, datalen, "./06_databases.out");
  
  *dleng = datalen;
  
  return data;
}

static char *
CreateListPlaylists (int * dleng)
{
  char * data;
  int datalen;
  int datamax;
  
  if (NULL == (data = (char *)malloc(sizeof(char) * 500))) {
    fprintf(stderr, "CreateListPlaylists.. malloc() failed [1]\n");
    exit(1);
  }
  
  datalen = 0;
  datamax = 200;
  
  bzero(data, datamax);
  
  datalen += WriteDAAPList    (data + datalen, "aply", NULL, 0);        // daap.databaseplaylists     response to /databases/<id>/containers
  datalen += WriteDAAPInt     (data + datalen, "mstt", 200);            // dmap.status                status code (200)
  datalen += WriteDAAPByte    (data + datalen, "muty", 0);              // dmap.updatetype            <no description available>
  datalen += WriteDAAPInt     (data + datalen, "mtco", 1);              // dmap.specifiedtotalcount   number of items in response to a request
  datalen += WriteDAAPInt     (data + datalen, "mrco", 1);              // dmap.returnedcount         number of items returned in a request
  datalen += WriteDAAPList    (data + datalen, "mlcl", NULL, 0);        // dmap.listing               a list
  datalen += WriteDAAPList    (data + datalen, "mlit", NULL, 0);        // dmap.listingitem           a single item in said list

  datalen += WriteDAAPInt     (data + datalen, "miid", 0x01);           // dmap.itemid                the playlist's item id
  datalen += WriteDAAPLong    (data + datalen, "mper", 0x1122LL);       // dmap.persistentid          the playlist's persistent id
  datalen += WriteDAAPString  (data + datalen, "minm", "Playlist", 8);  // dmap.itemname              the playlist's name
  datalen += WriteDAAPInt     (data + datalen, "mimc", 0x01);           // dmap.itemcount             the number of items in the playlist
  datalen += WriteDAAPByte    (data + datalen, "aeSP", 0x00);           // com.apple.itunes.smart-playlist

  WriteDAAPNameAndLength(data,      "aply", datalen - 8);
  WriteDAAPNameAndLength(data + 53, "mlcl", datalen - (53 + 8));
  WriteDAAPNameAndLength(data + 61, "mlit", datalen - (61 + 8));
  
  WriteDAAPFile(data, datalen, "./07_databases.out");
  
  *dleng = datalen;
  
  return data;
}

static char *
CreateGetPlaylist (int * dleng)
{
  char * data;
  int datalen;
  int datamax;
  
  if (NULL == (data = (char *)malloc(sizeof(char) * 500))) {
    fprintf(stderr, "CreateGetPlaylist.. malloc() failed [1]\n");
    exit(1);
  }
  
  datalen = 0;
  datamax = 500;
  
  bzero(data, datamax);
  
  datalen += WriteDAAPList    (data + datalen, "apso", NULL, 0);        // daap.playlistsongs         response to /databases/<id>/containers/<id>/items
  datalen += WriteDAAPInt     (data + datalen, "mstt", 200);            // dmap.status                status code (200)
  datalen += WriteDAAPByte    (data + datalen, "muty", 0);              // dmap.updatetype            <no description available>
  datalen += WriteDAAPInt     (data + datalen, "mtco", 1);              // dmap.specifiedtotalcount   number of items in response to a request
  datalen += WriteDAAPInt     (data + datalen, "mrco", 1);              // dmap.returnedcount         number of items returned in a request
  datalen += WriteDAAPList    (data + datalen, "mlcl", NULL, 0);        // dmap.listing               a list
  datalen += WriteDAAPList    (data + datalen, "mlit", NULL, 0);        // dmap.listingitem           a single item in said list

  datalen += WriteDAAPByte    (data + datalen, "mikd", 0x02);           // dmap.itemkind              type '2' for music
  datalen += WriteDAAPInt     (data + datalen, "miid", 0x01);           // dmap.itemid                item id of the song
  datalen += WriteDAAPInt     (data + datalen, "mcti", 0x01);           // dmap.containeritemid       id in the container
  
  WriteDAAPNameAndLength(data,      "apso", datalen - 8);
  WriteDAAPNameAndLength(data + 53, "mlcl", datalen - (53 + 8));
  WriteDAAPNameAndLength(data + 61, "mlit", datalen - (61 + 8));
  
  WriteDAAPFile(data, datalen, "./08_databases.out");
  
  *dleng = datalen;
  
  return data;
}

static char *
CreateStreamSong (int * dleng)
{
  char * data;
  int datalen;
  int datamax;
  
  if (NULL == (data = (char *)malloc(sizeof(char) * 6181518))) {
    fprintf(stderr, "CreateGetPlaylist.. malloc() failed [1]\n");
    exit(1);
  }
  
  datalen = 0;
  datamax = 500;
  
  bzero(data, datamax);

  FILE * file;
  
  if (NULL == (file = fopen("./song.mp3", "r"))) {
    fprintf(stderr, "CreateStreamSong.. fopen() failed\n");
    exit(1);
  }
  
  datalen = fread(data, sizeof(char), 6181518, file);
  
  /**
  datalen += WriteDAAPList    (data + datalen, "apso", NULL, 0);        // daap.playlistsongs         response to /databases/<id>/containers/<id>/items
  datalen += WriteDAAPInt     (data + datalen, "mstt", 200);            // dmap.status                status code (200)
  datalen += WriteDAAPByte    (data + datalen, "muty", 0);              // dmap.updatetype            <no description available>
  datalen += WriteDAAPInt     (data + datalen, "mtco", 1);              // dmap.specifiedtotalcount   number of items in response to a request
  datalen += WriteDAAPInt     (data + datalen, "mrco", 1);              // dmap.returnedcount         number of items returned in a request
  datalen += WriteDAAPList    (data + datalen, "mlcl", NULL, 0);        // dmap.listing               a list
  datalen += WriteDAAPList    (data + datalen, "mlit", NULL, 0);        // dmap.listingitem           a single item in said list
  
  datalen += WriteDAAPByte    (data + datalen, "mikd", 0x02);           // dmap.itemkind              type '2' for music
  datalen += WriteDAAPInt     (data + datalen, "miid", 0x01);           // dmap.itemid                item id of the song
  datalen += WriteDAAPInt     (data + datalen, "mcti", 0x01);           // dmap.containeritemid       id in the container
  
  WriteDAAPNameAndLength(data,      "apso", datalen - 8);
  WriteDAAPNameAndLength(data + 53, "mlcl", datalen - (53 + 8));
  WriteDAAPNameAndLength(data + 61, "mlit", datalen - (61 + 8));
  **/
  
  WriteDAAPFile(data, datalen, "./10_databases.out");
  
  *dleng = datalen;
  
  return data;
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Listener
#endif

static int        gServerSock = 0;
static pthread_t *gServerThread;
static struct sockaddr_in *gServerAddr;

static int
ReadLine (Conn * conn)
{
  int i = 0;
  
  for (i = 0; i < conn->buflen; i++)
    if (*(conn->readbuf + i) == '\r' && *(conn->readbuf + i + 1) == '\n')
      return i + 2;
  
  return -1;
}

// Here is an example of what iTunes v4.2 (72) sends as the first step
//
//   GET /server-info HTTP/1.1
//   Host: 10.0.1.70:5090
//   Accept: */*
//   User-Agent: iTunes/4.2 (Macintosh; N; PPC)
//   Client-DAAP-Version: 2.0
//   Client-DAAP-Access-Index: 1
//   Accept-Encoding: gzip
//
static void
ProcessHttpServerInfoRequest (Conn * conn)
{
  fprintf(stderr, "ProcessHttpServerInfoRequest..\n");
  
  int datalen = 0;
  int replylen = 0;
  char * data;
  
  if (NULL == (conn->http->http_reply = (char *)malloc(sizeof(char) * 500))) {
    fprintf(stderr, "ProcessHttpServerInfoRequest.. failed to malloc() for http_reply\n");
    exit(1);
  }

  data = CreateServerInfo(&datalen);
  
  memcpy(conn->http->http_reply + replylen, "HTTP/1.1 200 OK\r\n",                         17); replylen += 17;
  memcpy(conn->http->http_reply + replylen, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",     37); replylen += 37;
  memcpy(conn->http->http_reply + replylen, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",      36); replylen += 36;
  memcpy(conn->http->http_reply + replylen, "Content-Type: application/x-dmap-tagged\r\n", 41); replylen += 41;
  memcpy(conn->http->http_reply + replylen, "Content-Length: ",                            16); replylen += 16;
  replylen += sprintf(conn->http->http_reply + replylen, "%d", datalen);
  memcpy(conn->http->http_reply + replylen, "\r\n\r\n",                                     4); replylen +=  4;
  
  memcpy(conn->http->http_reply + replylen, data, datalen); replylen += datalen;

  WriteDAAPFile(conn->http->http_reply, replylen, "./01_server-info.http");
  
  write(conn->sock, conn->http->http_reply, replylen);
  
  free(data);
  free(conn->http->http_reply);
}

// Here is an example of what iTunes v4.2 (72) sends as the second step:
//
//   GET /content-codes HTTP/1.1
//   Host: 10.0.1.70
//   Client-DAAP-Version: 2.0
//   User-Agent: iTunes/4.2 (Macintosh; N; PPC)
//   Accept-Language: en-us, en;q=0.50
//   Client-DAAP-Access-Index: 1
//   Accept-Encoding: gzip
//
static void
ProcessHttpContentCodesRequest (Conn * conn)
{
  int datalen = 0;
  int replylen = 0;
  char * data;
  
  if (NULL == (conn->http->http_reply = (char *)malloc(sizeof(char) * 10500))) {
    fprintf(stderr, "ProcessHttpServerInfoRequest.. failed to malloc() for http_reply\n");
    exit(1);
  }
  
  data = CreateContentCodes(&datalen);
  
  memcpy(conn->http->http_reply + replylen, "HTTP/1.1 200 OK\r\n",                         17); replylen += 17;
  memcpy(conn->http->http_reply + replylen, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",     37); replylen += 37;
  memcpy(conn->http->http_reply + replylen, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",      36); replylen += 36;
  memcpy(conn->http->http_reply + replylen, "Content-Type: application/x-dmap-tagged\r\n", 41); replylen += 41;
  memcpy(conn->http->http_reply + replylen, "Content-Length: ",                            16); replylen += 16;
  replylen += sprintf(conn->http->http_reply + replylen, "%d", datalen);
  memcpy(conn->http->http_reply + replylen, "\r\n\r\n",                                     4); replylen +=  4;
  
  memcpy(conn->http->http_reply + replylen, data, datalen); replylen += datalen;
  
  WriteDAAPFile(conn->http->http_reply, replylen, "./02_content-codes.http");
  
  write(conn->sock, conn->http->http_reply, replylen);
  
  free(data);
  free(conn->http->http_reply);
}

// Here is an example of what iTunes v4.2 (72) sends as the third step:
//
//   GET /login HTTP/1.1
//   Host: 10.0.1.70
//   Client-DAAP-Version: 2.0
//   User-Agent: iTunes/4.2 (Macintosh; N; PPC)
//   Accept-Language: en-us, en;q=0.50
//   Client-DAAP-Access-Index: 1
//   Accept-Encoding: gzip
//
static void
ProcessHttpLoginRequest (Conn * conn)
{
  int datalen = 0;
  int replylen = 0;
  char * data;
  
  if (NULL == (conn->http->http_reply = (char *)malloc(sizeof(char) * 500))) {
    fprintf(stderr, "ProcessHttpServerInfoRequest.. failed to malloc() for http_reply\n");
    exit(1);
  }
  
  data = CreateLogin(&datalen);
  
  memcpy(conn->http->http_reply + replylen, "HTTP/1.1 200 OK\r\n",                         17); replylen += 17;
  memcpy(conn->http->http_reply + replylen, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",     37); replylen += 37;
  memcpy(conn->http->http_reply + replylen, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",      36); replylen += 36;
  memcpy(conn->http->http_reply + replylen, "Content-Type: application/x-dmap-tagged\r\n", 41); replylen += 41;
  memcpy(conn->http->http_reply + replylen, "Content-Length: ",                            16); replylen += 16;
  replylen += sprintf(conn->http->http_reply + replylen, "%d", datalen);
  memcpy(conn->http->http_reply + replylen, "\r\n\r\n",                                     4); replylen +=  4;
  
  memcpy(conn->http->http_reply + replylen, data, datalen); replylen += datalen;
  
  WriteDAAPFile(conn->http->http_reply, replylen, "./03_login.http");
  
  write(conn->sock, conn->http->http_reply, replylen);
  
  free(data);
  free(conn->http->http_reply);
}

// Here is an example of what iTunes v4.2 (72) sends as the fourth step:
//
//   GET /update?session-id=123&revision-number=1 HTTP/1.1
//   Host: 10.0.1.70
//   Client-DAAP-Version: 2.0
//   User-Agent: iTunes/4.2 (Macintosh; N; PPC)
//   Accept-Language: en-us, en;q=0.50
//   Client-DAAP-Validation: 718325E416DF3FB2C8805B334101A6B5
//   Client-DAAP-Access-Index: 1
//   Accept-Encoding: gzip
//
static void
ProcessHttpUpdateRequest (Conn * conn)
{
  int datalen = 0;
  int replylen = 0;
  char * data;
  
  if (NULL == (conn->http->http_reply = (char *)malloc(sizeof(char) * 500))) {
    fprintf(stderr, "ProcessHttpServerInfoRequest.. failed to malloc() for http_reply\n");
    exit(1);
  }

  if (0 == strncmp((char *)&conn->http->cgi_query, "session-id=4660&revision-number=3&delta=3", 41)) {
    free(conn->http->http_reply);
    return;
  }
  else
    data = CreateUpdate(&datalen);
  
  memcpy(conn->http->http_reply + replylen, "HTTP/1.1 200 OK\r\n",                         17); replylen += 17;
  memcpy(conn->http->http_reply + replylen, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",     37); replylen += 37;
  memcpy(conn->http->http_reply + replylen, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",      36); replylen += 36;
  memcpy(conn->http->http_reply + replylen, "Content-Type: application/x-dmap-tagged\r\n", 41); replylen += 41;
  memcpy(conn->http->http_reply + replylen, "Content-Length: ",                            16); replylen += 16;
  replylen += sprintf(conn->http->http_reply + replylen, "%d", datalen);
  memcpy(conn->http->http_reply + replylen, "\r\n\r\n",                                     4); replylen +=  4;
  
  memcpy(conn->http->http_reply + replylen, data, datalen); replylen += datalen;
  
  if (0 == strncmp((char *)&conn->http->cgi_query, "session-id=4660&revision-number=3&delta=3", 41))
    WriteDAAPFile(conn->http->http_reply, replylen, "./09_update.http");
  else
    WriteDAAPFile(conn->http->http_reply, replylen, "./04_update.http");
  
  write(conn->sock, conn->http->http_reply, replylen);
  
  free(data);
  free(conn->http->http_reply);
}

// Here is an example of what iTunes v4.2 (72) sends as the fifth step:
//
//   GET /databases?session-id=4660&revision-number=3 HTTP/1.1
//   Host: 10.0.1.70
//   Client-DAAP-Version: 2.0
//   User-Agent: iTunes/4.2 (Macintosh; N; PPC)
//   Accept-Language: en-us, en;q=0.50
//   Client-DAAP-Validation: D548C2086370D5AD74B063820455CB73
//   Client-DAAP-Access-Index: 1
//   Accept-Encoding: gzip
//
//
// Here is an example of what iTunes v4.2 (72) sends as the sixth step:
//
//   GET /databases/32/items?type=music&meta=dmap.itemid,dmap.itemname,dmap.itemkind,
//   dmap.persistentid,daap.songalbum,daap.songgrouping,daap.songartist,daap.songbitrate,
//   daap.songbeatsperminute,daap.songcomment,daap.songcompilation,daap.songcomposer,
//   daap.songdateadded,daap.songdatemodified,daap.songdisccount,daap.songdiscnumber,
//   daap.songdisabled,daap.songeqpreset,daap.songformat,daap.songgenre,daap.songdescription,
//   daap.songrelativevolume,daap.songsamplerate,daap.songsize,daap.songstarttime,
//   daap.songstoptime,daap.songtime,daap.songtrackcount,daap.songtracknumber,
//   daap.songuserrating,daap.songyear,daap.songdatakind,daap.songdataurl,
//   com.apple.itunes.norm-volume&session-id=4660&revision-number=3 HTTP/1.1
//   Host: 10.0.1.70
//   Client-DAAP-Version: 2.0
//   User-Agent: iTunes/4.2 (Macintosh; N; PPC)
//   Accept-Language: en-us, en;q=0.50
//   Client-DAAP-Validation: 08D3152FA6F1D665396F02F52A664B8D
//   Client-DAAP-Access-Index: 1
//   Accept-Encoding: gzip
//
//
// Here is an example of what iTunes v4.2 (72) sends as the ninth step:
//
//   GET /databases/32/items/1.mp3?session-id=4660 HTTP/1.1
//   Host: 10.0.1.69:5090
//   Cache-Control: no-cache
//   Accept: */*
//   x-audiocast-udpport:49153
//   icy-metadata:1
//   User-Agent: iTunes/4.2 (Macintosh; N; PPC)
//   Connection: close
//   Client-DAAP-Access-Index: 1
//   Client-DAAP-Validation: FA4389E1814942AD2B46C123799B6EE1
//
static void
ProcessHttpDatabasesRequest (Conn * conn)
{
  fprintf(stderr, "ProcessHttpDatabasesRequest..\n");
  
  int datalen = 0;
  int replylen = 0;
  char * data;
  
  if (NULL == (conn->http->http_reply = (char *)malloc(sizeof(char) * 10000000))) {
    fprintf(stderr, "ProcessHttpDatabasesRequest.. failed to malloc() for http_reply\n");
    exit(1);
  }

  if (0 == strncmp((char *)&conn->http->file_path, "/databases/32/containers/1/items", 32))
    data = CreateGetPlaylist(&datalen);
  else if (0 == strncmp((char *)&conn->http->file_path, "/databases/32/items/1.mp3", 25))
    data = CreateStreamSong(&datalen);
  else if (0 == strncmp((char *)&conn->http->file_path, "/databases/32/containers", 24))
    data = CreateListPlaylists(&datalen);
  else if (0 == strncmp((char *)&conn->http->file_path, "/databases/", 11))
    data = CreateListSongs(&datalen);
  else
    data = CreateDatabasesList(&datalen);
  
  memcpy(conn->http->http_reply + replylen, "HTTP/1.1 200 OK\r\n",                         17); replylen += 17;
  memcpy(conn->http->http_reply + replylen, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",     37); replylen += 37;
  memcpy(conn->http->http_reply + replylen, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",      36); replylen += 36;
  memcpy(conn->http->http_reply + replylen, "Content-Type: application/x-dmap-tagged\r\n", 41); replylen += 41;
  memcpy(conn->http->http_reply + replylen, "Content-Length: ",                            16); replylen += 16;
  replylen += sprintf(conn->http->http_reply + replylen, "%d", datalen);
  memcpy(conn->http->http_reply + replylen, "\r\n\r\n",                                     4); replylen +=  4;
  
  memcpy(conn->http->http_reply + replylen, data, datalen); replylen += datalen;
  
  if (0 == strncmp((char *)&conn->http->file_path, "/databases/32/containers/1/items", 32))
    WriteDAAPFile(conn->http->http_reply, replylen, "./08_databases.http");
  else if (0 == strncmp((char *)&conn->http->file_path, "/databases/32/items/1.mp3", 25))
    WriteDAAPFile(conn->http->http_reply, replylen, "./10_databases.http");
  else if (0 == strncmp((char *)&conn->http->file_path, "/databases/32/containers", 24))
    WriteDAAPFile(conn->http->http_reply, replylen, "./07_databases.http");
  else if (0 == strncmp((char *)&conn->http->file_path, "/databases/", 11))
    WriteDAAPFile(conn->http->http_reply, replylen, "./06_databases.http");
  else
    WriteDAAPFile(conn->http->http_reply, replylen, "./05_databases.http");
  
  write(conn->sock, conn->http->http_reply, replylen);
  
  free(data);
  free(conn->http->http_reply);
}

static void
ProcessHttpFileNotFound (Conn * conn)
{
  
}

static void
ProcessHttpRequest (Conn * conn)
{
  if (0 == strcmp((char *)&conn->http->file_path, "/server-info"))
    ProcessHttpServerInfoRequest(conn);
  else if (0 == strcmp((char *)&conn->http->file_path, "/content-codes"))
    ProcessHttpContentCodesRequest(conn);
  else if (0 == strcmp((char *)&conn->http->file_path, "/login"))
    ProcessHttpLoginRequest(conn);
  else if (0 == strcmp((char *)&conn->http->file_path, "/update"))
    ProcessHttpUpdateRequest(conn);
  else if (0 == strncmp((char *)&conn->http->file_path, "/databases", 10))
    ProcessHttpDatabasesRequest(conn);
//else if (0 == strcmp((char *)&conn->http->file_path, "/db/id/items"))
//  ProcessHttpSomethingRequest(conn);
  else
    ProcessHttpFileNotFound(conn);
}

static void
ParseGetRequest (Conn * conn, int linepos)
{
  int i = 4;
  int j = 0;
  
  bzero(&conn->http->file_path, 1000);
  bzero(&conn->http->cgi_query, 1000);
  
  // copy from the '/' following GET until (but not including) the first space
  // or the first question mark (which would begin the parameters section)
  for (i = 4; i < linepos - 2; i++) {
    if (*(conn->readbuf + i) == ' ' || *(conn->readbuf + i) == '?') {
      memcpy(&conn->http->file_path, (conn->readbuf + 4), i - 4);
      break;
    }
  }

  // copy from the '?' to the first space
  if (*(conn->readbuf + i) == '?') {
    for (j = i + 1; j < linepos - 2; j++) {
      if (*(conn->readbuf + j) == ' ') {
        memcpy(&conn->http->cgi_query, (conn->readbuf + i + 1), j - (i+1));
        break;
      }
    }
  }
      
  fprintf(stderr, "ParseGetRequest.. GET request: '%s'\n", (char *)&conn->http->file_path);
  fprintf(stderr, "ParseGetRequest.. GET query  : '%s'\n", (char *)&conn->http->cgi_query);
}

static void *
ConnectionLoop (void *arg)
{
  fprintf(stderr, "ConnectionLoop.. starting\n");

  Conn * conn;
  int nfds = 0;
  fd_set readfds;
  int result;

  conn = (Conn *)arg;
  
  FD_ZERO(&readfds);
  FD_SET(conn->sock, &readfds);
  
  nfds = conn->sock + 1;

  if (NULL == (conn->readbuf = (char *)malloc(sizeof(char) * 2000))) {
    fprintf(stderr, "ConnectionLoop.. malloc() failed for readbuf\n");
    exit(1);
  }
  
  if (NULL == (conn->http = (Http *)malloc(sizeof(Http)))) {
    fprintf(stderr, "ConnectionLoop.. malloc() failed for http\n");
    exit(1);
  }

  conn->buflen = 0;
  conn->bufmax = 2000;
  bzero(conn->readbuf, conn->bufmax);
  bzero(&conn->http->file_path, 1000);
  bzero(&conn->http->cgi_query, 1000);
  
  while (!gHault)
  {
    int linepos = 0;
    char * linedat;
    
    fprintf(stderr, "ConnectionLoop.. waiting on select()\n");
    
    if (0 >= (result = select(nfds, &readfds, NULL, NULL, NULL))) {
      fprintf(stderr, "ConnectionLoop.. select() failed with, %d\n", result);
      break;
    }
    
    fprintf(stderr, "ConnectionLoop.. select() returned with, %d\n", result);

    if (0 >= (result = read(conn->sock, conn->readbuf, conn->bufmax))) {
      fprintf(stderr, "ConnectionLoop.. read() failed\n");
      break;
    }
    
    conn->buflen = result;

    fprintf(stderr, "ConnectionLoop.. read() returned %d bytes\n", result);
  
//  int i = 0;
//  for (i = 0; i < conn->buflen; i++)
//    fprintf(stderr, "%c", *(conn->readbuf + i));
    
    while (-1 != (linepos = ReadLine(conn))) {
      if (linepos == 2) {
        fprintf(stderr, "ConnectionLoop.. calling ProcessHttpRequest()\n");
        ProcessHttpRequest(conn);
        bzero(&conn->http->file_path, 1000);
        bzero(&conn->http->cgi_query, 1000);
        break;
      }
      else if (0 == strncmp(conn->readbuf, "GET /", 5)) {
        ParseGetRequest(conn, linepos);
      }
      
      linedat = (char *)malloc(sizeof(char) * (linepos-1));
      bzero(linedat, linepos - 1);
      memcpy(linedat, conn->readbuf, (linepos-2));
      fprintf(stdout, "DATA [%d]: '%s'\n", (linepos-2), linedat);
      free(linedat);

      // shift the remaining data in the buffer down to the beginning of the buffer
      memcpy(conn->readbuf, conn->readbuf + linepos, conn->buflen - linepos);
      conn->buflen -= linepos;
    }
  }
  
  free(conn->readbuf);
  free(conn->http);
  free(conn);
  
  return NULL;
}

static void *
ListenerLoop (void *arg)
{
  fprintf(stderr, "ListenerLoop.. starting\n");

  mStatus status = mStatus_NoError;
  Conn * conn;

  // we malloc() a Conn here for holding the data associated with a new client
  // connection. we only create a new one after a new connection has been
  // successfully spun off into a thread ... otherwise, we reuse it.
  if (NULL == (conn = (Conn *)malloc(sizeof(Conn)))) {
    fprintf(stderr, "ListenerLoop.. malloc() failed for Conn\n");
    exit(1);
  }
  
  while (!gHault)
  {
    if (0 >= (conn->sock = accept(gServerSock, (struct sockaddr *)&conn->addr, (int *)&conn->addrlen))) {
      fprintf(stderr, "ListenerLoop.. accept() failed.\n");
      continue;
    }
    
    fprintf(stderr, "ListenerLoop.. accepted new connection\n");
    
    if (0 != (status = pthread_create(&conn->thread, NULL, ConnectionLoop, conn))) {
      fprintf(stderr, "ListenerLoop..pthread_create() failed for new thread\n");
      continue;
    }

    // malloc() another Conn for the next incoming connection
    if (NULL == (conn = (Conn *)malloc(sizeof(Conn)))) {
      fprintf(stderr, "ListenerLoop.. malloc() failed for Conn\n");
      exit(1);
    }
  }
  
  return NULL;
}

static void
StartListener (void)
{
  mStatus status;
  
  if (0 >= (gServerSock = socket(AF_INET, SOCK_STREAM, 0))) {
    fprintf(stderr, "StartListener.. socket() failed, %d\n", gServerSock);
    exit(1);
  }

//{
//  struct timeval sotimeout;
//  sotimeout.tv_sec = 100;
//  sotimeout.tv_usec = 0;
//  
//  setsockopt(gServerSock, SOL_SOCKET, SO_RCVTIMEO, &sotimeout, sizeof(struct timeval));
//}
  
  if (NULL == (gServerAddr = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in)))) {
    fprintf(stderr, "StartListener.. malloc() failed\n");
    exit(1);
  }

  bzero(gServerAddr, sizeof(struct sockaddr_in));

  gServerAddr->sin_family = AF_INET;
  gServerAddr->sin_addr.s_addr = htonl(INADDR_ANY);
  gServerAddr->sin_port = htons(kDefaultPortNumber);
  
  if (0 != (status = bind(gServerSock, (struct sockaddr *)gServerAddr, sizeof(struct sockaddr)))) {
    fprintf(stderr, "StartListener.. bind() failed, %d\n", (int)status);
    exit(1);
  }
  
  if (0 != (status = listen(gServerSock, 16))) {
    fprintf(stderr, "StartListener.. listen() failed\n");
    exit(1);
  }
  
  if (NULL == (gServerThread = (pthread_t *)malloc(sizeof(pthread_t)))) {
    fprintf(stderr, "StartListener.. malloc(pthread_t) failed\n");
    exit(1);
  }
  
  bzero(gServerThread, sizeof(pthread_t));
  
  if (0 != (status = pthread_create(gServerThread, NULL, ListenerLoop, NULL))) {
    fprintf(stderr, "StartListener.. pthread_create() failed, %ld\n", status);
    exit(1);
  }

  fprintf(stderr, "StartingListener.. success!\n");
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Command Line Arguments
#endif

static const char *gServiceRoot     = kDefaultServiceRoot;
static const char *gServicePass     = kDefaultServicePass;
static const char *gServiceName     = "iCast";
static const char *gServiceType     = kDefaultServiceType;
static const char *gServiceDomain   = kDefaultServiceDomain;
static const char *gPIDFile         = kDefaultPIDFile;
static int         gPortNumber      = kDefaultPortNumber;
static mDNSu16     gServiceTextLen  = 0;
static mDNSu8      gServiceText[sizeof(RDataBody)];

static void
ParseArguments (int argc, char **argv)
{
  int ch;
  
  // Set gProgramName to the last path component of argv[0]
  
  if (NULL == (gProgramName = strrchr(argv[0], '/')))
    gProgramName = argv[0];
  else
    gProgramName += 1;

  // Parse command line options using getopt
  
  do {
    if (-1 == (ch = getopt(argc, argv, "")))
      continue;
    
    switch (ch) {
      case 'p':
        gServicePass = optarg;
        break;
      case 'r':
        gServiceRoot = optarg;
        break;
      default:
        fprintf(stderr, "Unknown argument, %c\n", ch);
        exit(1);
        break;
    }
  } while (ch != -1);
  
  // gMDNSPlatformPosixVerboseLevel = 2;
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Registration
#endif

typedef struct PosixService PosixService;

struct PosixService {
  ServiceRecordSet coreServ;
  PosixService *next;
  int serviceID;
};

static PosixService *gServiceList = NULL;

static int gServiceID = 0;

static void
RegistrationCallback (mDNS * const m, ServiceRecordSet * const thisRegistration, mStatus status)
{
  (void)m;
  (void)thisRegistration;
  (void)status;
  
  fprintf(stderr, "RegistrationCallback().. I'm here.\n");
}

static mStatus
RegisterOneService (const char * serviceName,
                    const char * serviceType,
                    const char * serviceDomain,
                    const mDNSu8 text[],
                    mDNSu16      textLen,
                    long         portNumber)
{
  mStatus         status;
  PosixService *  thisServ;
  mDNSOpaque16    port;
  domainlabel     name;
  domainname      type;
  domainname      domain;
  
  if (NULL == (thisServ = (PosixService *)malloc(sizeof(*thisServ))))
    return mStatus_NoMemoryErr;
  
  MakeDomainLabelFromLiteralString(&name, serviceName);
  MakeDomainNameFromDNSNameString(&type, serviceType);
  MakeDomainNameFromDNSNameString(&domain, serviceDomain);
  
  port.b[0] = (portNumber >> 8) & 0x0FF;
  port.b[1] = (portNumber >> 0) & 0x0FF;
  
  status = mDNS_RegisterService (
    &mDNSStorage,                       // where the core stores its stuff
    &thisServ->coreServ,                // something important, i'm sure
    &name, &type, &domain,              // name, type, domain
    NULL, port,                         // host and port
    text, textLen,                      // txt data, length
    NULL, 0,                            // subtypes
    mDNSInterface_Any,                  // interface id
    RegistrationCallback, thisServ);    // callback and context
  
  if (status != mStatus_NoError) {
    if (thisServ != NULL) free(thisServ);
    return status;
  }
  
  thisServ->serviceID = gServiceID;
  gServiceID += 1;
  
  thisServ->next = gServiceList;
  gServiceList = thisServ;
  
  fprintf(stderr, "%s: Registered service %d, name '%s', type '%s', port %ld\n",
          gProgramName,
          thisServ->serviceID,
          serviceName,
          serviceType,
          portNumber);
  
  return mStatus_NoError;
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Main
#endif

int
main (int argc, char **argv)
{
  mStatus status;
  int result;
  
  (void)result;
  (void)gPIDFile;
  
  ParseArguments(argc, argv);

  // daemon(0,0);
  
  status = mDNS_Init (
    &mDNSStorage,
    &PlatformStorage,
    mDNS_Init_NoCache,
    mDNS_Init_ZeroCacheSize,
    mDNS_Init_AdvertiseLocalAddresses,
    mDNS_Init_NoInitCallback,
    mDNS_Init_NoInitCallbackContext);

  if (mStatus_NoError != status)
    return(2);

  RegisterOneService (
    gServiceName,
    gServiceType,
    gServiceDomain,
    gServiceText,
    gServiceTextLen,
    gPortNumber);
  
  signal(SIGHUP,   HandleSigHup);      // SIGHUP has to be sent by kill -HUP <pid>
  signal(SIGINT,   HandleSigInt);      // SIGINT is what you get for a Ctrl-C
  signal(SIGQUIT,  HandleSigQuit);     // SIGQUIT is what you get for a Ctrl-
  signal(SIGUSR1,  HandleSigUsr1);     // SIGUSR1 has to be sent by kill -USR1 <pid>
  signal(SIGUSR2,  HandleSigUsr2);     // SIGUSR2 has to be sent by kill -USR2 <pid>

  int datalen, datamax;
  
  StartListener();
//QueryiTunes();
//CreateServerInfo(&datalen, &datamax);
//CreateServerInfo();
  
  while (!gHault)
  {
    // fprintf(stderr, "Sitting in main loop\n");
    
    int nfds = 0;
    fd_set readfds;
    struct timeval timeout;
    int result;

    FD_ZERO(&readfds);
    
    timeout.tv_sec  = 0x00000001;
    timeout.tv_usec = 0x00;
    
    mDNSPosixGetFDSet(&mDNSStorage, &nfds, &readfds, &timeout);

    if (0 <= (result = select(nfds, &readfds, NULL, NULL, &timeout)))
      mDNSPosixProcessFDSet(&mDNSStorage, &readfds);
    else
      fprintf(stderr, "Error.. select() returned %d errno %d\n", result, errno);
  }

  fprintf(stderr, "Exiting.\n");
  
  mDNS_DeregisterService(&mDNSStorage, &gServiceList->coreServ);
  mDNS_Close(&mDNSStorage);
  
  return 0;
}
